# this file deals with the setup and interaction with the receiver thread
# interaction is done through a pipeline and a stack
# the stack is used to store the bytes of messages sent to this robot
# the pipline is used to give other instructions to the thread, things like settings updates
# both the stack and pipeline are one way, the stack is used only for sending back data to the main thread, the pipeline is used only for sending instructions to the reciever thread
#


from .sender import sendData
import yodel.globaldat as globaldat
from .classes import *
import yodel.standardformats as standardformats
from .dynamicheaders import *
from yodel.config import *
from typing import *


import sys
import multiprocessing as mp
incoming = mp.Queue()
globaldat.receiver_pipe, receiver_pipe_output = mp.Pipe()


# check to see if a given message is intended for this computer, either to
# receive or to relay
def is_recipient(
        data: bytearray, radiotap_header_length: int) -> Union[bool, Tuple[bool, bool]]:
    """
    check to see if a given message is intended for this computer and if the message should be relayed


    Args:
    	 data: raw bytes for incoming message

    	 radiotap_header_length: length of radiotap header, the header is skipped over because it does not hold yodel data.
    """

    # get data frame payload section
    frame = data[radiotap_header_length + 26 + 5:]
    pos = 0  # pos is used as a pointer to the current section of the header being decoded

    """
     message id, the semi unique identifer to each message to avoid receiving
     them twice
    """
    message_ID = frame[pos:pos + 4]
    if message_ID == globaldat.lastMid:  # since messages are repeated a lot it is worth saving the previous message id so that the array does not need to be fully indexed
        return (False)

    if message_ID not in globaldat.lastMessages:  # check if message has already been received

        globaldat.lastMid = message_ID  # set last mid to the current mid
        globaldat.lastMessages.append(message_ID)
        if len(globaldat.lastMessages) > 64:
            del globaldat.lastMessages[0]

        pos += 4
        namelen = globaldat.getInt(frame[pos:pos + 1])  # get length of name
        pos += 1

        name = frame[pos:pos + namelen].decode("utf-8")
        nameM = (name == globaldat.robotName or namelen == 0)

        pos += namelen
        gnamelen = globaldat.getInt(frame[pos:pos + 1])
        pos += 1
        group = frame[pos:pos + gnamelen].decode("utf-8")
        pos += gnamelen

        groupM = (group in globaldat.groups or gnamelen == 0)

        relay = (globaldat.relay == True and not (name == globaldat.robotName))

        return(nameM and groupM, relay)

    return (False)


def relayFrame(frame: bytearray) -> NoReturn:
    """
    used to allow receiver thread to relay messages back out

    Args:
    	 frame: bytes of message being relayed

    """

    header = b"\x72\x6f\x62\x6f\x74"  # yodel identifier
    if globaldat.relay == True:  # check if relay is enabled

        # format and pass data to sender thread
        sendData(header + frame, globaldat.maxRelay)


def listenrecv(pipe: mp.Pipe) -> bytearray:
    """
    listen for yodel messages, check if the are meant for this computer and if so receive them

    Args:
    	 pipe: same pipe as in receiver, used to check for settings updates
    """

    try:

        data = globaldat.yodelSocket.recv(globaldat.ETH_FRAME_LEN)

    except BaseException:
        return (None)

    radiolen = globaldat.getInt(data[2:3])
    fcs = data[17]  # fcs status flag is stored at byte 17
    if fcs == 2:  # if fcs flag is set then remove the fcs which is the last 4 bytes
        data = data[0:-4]

    payload = data[radiolen + 26:]

    pos = 0

    starth = (payload[pos:pos + 5])

    # radio tap headers are stripped on external frames (non loopback)
    if starth == b"\x72\x6f\x62\x6f\x74":
        
        # additionally check for settings change if any yodel data is captured
        # to see if anything important has changed
        settings_check(pipe)
        rdata = is_recipient(data, radiolen)

        if rdata:
            isr, dorelay = rdata
            if dorelay:
                relayFrame(payload[5:])
            if isr:

                return(payload[5:])

    return(None)


def settings_check(pipe) -> NoReturn:
    """ reads from pipe to detect settings updates

    Args:
    	 pipe: same settings pipe as used in receiver
    """
    if pipe.poll(0):  # check for new data in pipe
        settings = pipe.recv()  # get data
        #print(settings,"setting update")
        setting_update(settings[0], settings[1])  # change settings accordingly


def listen() -> Section:
    """
    used to listen for data being sent to your robot, is non-blocking

    Returns:
        yodel.Section: contains any received message which is accessible through the .payload attribute, also contain meta data about the message 
    """
    global incoming
    try:  # make specific error
        frame = incoming.get(False)
        data = frame.frame  # take bytes from message
        # take bytes and extract header info
        data = decode(data, standardformats.standard_header_format)
        return(data)
    except BaseException:

        return(None)


def receiver(incoming: mp.Queue, pipe: mp.Pipe) -> NoReturn:
    """
    this is the main function in the reciever thread. this checks for new messages and stores them into the stack so that main thread can access them
    this also checks for settings updates and acts on them


    Args:
    	 incoming: queue that main thread reads from to receive new messages

    	 pipe: settings updates
    """
    globaldat.yodelSocket.settimeout(.1)
    while True:

        settings_check(pipe)  # check for settings updates
        dat = listenrecv(pipe)

        if dat is not None:
            formatted = FrameRecv(dat)

            incoming.put(formatted)

            if incoming.full():
                incoming.get()



import multiprocessing
from multiprocessing import Pipe
from multiprocessing.queues import Queue
from socket import socket
from typing import *

lastMessages: List[int] = [] #holds the list of past message id's received, used to avoid receiving the same message twice
iface: str = ""  # interface name, set during runtime
robotName: str = ""  # robot name, set by setName during runtime
groups: List[str] = [""]  # list of groups bot is a part of
delay: int = 0
ETH_P_ALL: int = 3
ETH_FRAME_LEN: int = 1514  # Max. octets in frame sans FCS
relay: bool = False
maxRelay: int = 5
totalsends: int = 10
RADIO_TAP: bytearray = b"\x00\x00\x22\x00\xae\x40\x00\xa0\x20\x08\x00\xa0\x20\x08\x00\x00\x00\x10\x10\x02\x6c\x09\xa0\x00\xb0\x00\x64\x00\x00\x00\x00\x00\x00\x01"
lastMid: int = 0  # message ID of last message received

yodelSocket: socket = None
debug: bool = True
sender_thread: multiprocessing.Process = None
receiver_thread: multiprocessing.Process = None
receiver_pipe: Pipe = None
sender_pipe: Pipe = None
outgoing: Queue = None

messages_types: List[int] = [0] * 256

# how much time in seconds the threads have to finish what they are doing
# once the program ends
SOCKET_CLOSE_TIME: float = .1


def getInt(bytea: bytearray) -> int:
    return (int.from_bytes(bytea, byteorder='little'))


def bytesPrint(x: bytearray) -> NoReturn:
    print(''.join(r'\x' + hex(letter)[2:] for letter in x))



"""
this file defines some of the formats used internally for wrapping various types of messages the user might send



"""



from yodel.dynamicheaders import *


standard_header = [  # begins all messages
    Field("mid", int, bytes=4),  # message id
    Field("Rname", str, bytes=255),  # receiver name
    Field("Gname", str, bytes=255),  # group identifier
    Field("Sname", str, bytes=255),  # sender name
    Field("mtype", int, bytes=1)  # message type
]


request = [  # mtype: 1
    Field("request", str, 255),
]
response = [  # mtype: 2
    Field("response", str, 255),
]


standard_header_format = Format(standard_header)



# this file deals with the setup and interaction with the sender thread
# interaction is done through a pipeline and a stack
# the stack is used to give the sender thread data to send out
# the pipline is used to give other instructions to the thread, things like settings updates
# both the stack and the pipeline are assumed to be one-way, that being from the main thread to the sender thread
#

import multiprocessing as mp
import yodel.standardformats as standardformats
from .dynamicheaders import *
import random
from .classes import *
from yodel.config import *
import sys
import yodel.globaldat as globaldat
from yodel.config import *
from typing import *

# setup for communication with sender thread
# print("init")
globaldat.sender_pipe, sender_pipe_output = mp.Pipe()
globaldat.outgoing = mp.Queue()
#####


# template that gets filled whenever send is called. this is to avoid
# having to generate a new class object every time send is called
outgoing_data = Section(standardformats.standard_header_format)
exiting = False
exist_start = 0


def sendData(packet: FrameStruct, repeats: int) -> NoReturn:
    """
    Generate 80211 header and send completed data

    Args:
     	packet: framestruct that holds raw outgoing data as well and info about how it should be sent

     	repeats: number of times message should be sent
    """
    ftype = b'\x08\x00'
    dur = b'\x00\x00'
    # random hex stream, could be used as additional space of bits
    src = b'\x08\x00\x27\x8e\x75\x44'
    # broadcast address is used to stop certain drivers retransmitting frames
    dst = b'\xff\xff\xff\xff\xff\xff'
    bssid = src
    # semi unique id, annoyingly not usable due to lack of bits for this appli
    sn = (random.randint(0, 4096))
    sn = sn << 4
    seq = sn.to_bytes(4, 'little')

    # generate 80211 header
    header80211 = ftype + dur + dst + src + bssid + seq

    # combine header with other data to create valid frame
    data = globaldat.RADIO_TAP + header80211 + b"\x72\x6f\x62\x6f\x74" + \
        packet  # attach radiotap headers, 80211 headers and yodel payload

    for i in range(repeats):  # re-transmmit message a couple times
        globaldat.yodelSocket.send(data)  # send the data


def send(payload: any, name: str = "", group: str = "") -> NoReturn:
    """
    take in payload and generate additional data needed to be complient with yodel standard header and add it to the stack for the thread to access.

    Args:
        payload: data being sent

        name: name of recipient

        group: group of recipient
    """

    global outgoing, outgoing_data

    # name = kwargs.get("name", '')    #receiver name
    # group = kwargs.get("group", '')  #receiver group
    mtype = 0  # kwargs.get("type", '')   #message type

    if isinstance(
            payload, Section):  # if type is a section than it can be processed automatically
        mtype = payload.format.mtype
        payload = bytes(payload)

    if name:  # check for a provided receiver name otherwise make it blank
        outgoing_data.Rname = name
    else:
        outgoing_data.Rname = ""

    if group:  # check for a provided receiving group, otherwise make it blank
        outgoing_data.Gname = group
    else:
        outgoing_data.Gname = ""

    # set the Sender name for the outgoing data to be equal to the robots name
    outgoing_data.Sname = globaldat.robotName
    # generate random indetifier for the message
    outgoing_data.mid = random.randint(-2147483648, 2147483647)
    # take the payload and convert it to bytes
    outgoing_data.payload = typeManagment(payload)
    outgoing_data.mtype = mtype

    encoded_frame = bytes(outgoing_data)  # get bytes
    # create object that holds raw bytes as well as data about how it should
    # be sent
    outgoing_frame = FrameStruct(encoded_frame)
    outgoing_frame.repeats = globaldat.totalsends
    # push data onto stack so that sender thread can access and send off frame
    globaldat.outgoing.put(outgoing_frame)


def sender(outgoing: mp.Queue, pipe: mp.Pipe) -> NoReturn:  # thread that manages sending out data
    """
    create thread to manage sending data from outgoing stack
    
    Args:
    	 outgoing: queue that holds all messages being sent

    	 pipe: used to send settings updates and other status updates from main to this thread

    """
    while True:
        if pipe.poll(
                0):  # check for any new settings updates or other instructions from the main thread

            settings = pipe.recv()  # if there are any then receive them
            # use these as inputs to the settings update function
            setting_update(settings[0], settings[1])
        frame = outgoing.get()  # wait for data in stack to be sent (is blocking)
        sendData(frame.bytes, frame.repeats)



import time


class FrameRecv:  # this is used to store messages in the stack that are recieved by the reciever thread
    def __init__(self, frame: bytearray):
        self.frame = frame  # raw message bytes
        self.time = time.time()  # time message is received


'''
class fmeta:
    ftype = 0
'''


class FrameStruct:
    relay = False
    ftype = 0
    sender = ""
    receiver = ""
    group = ""
    mid = 0
    payload = ""
    repeats = None

    def __repr__(self):
        return(self.payload)

    def __init__(self, frame: bytearray):
        self.bytes = frame


''''
class frameR:
    ftype = 0
    def __init__(self, frame, repeats):
        self.frame = frame
        self.repeats = repeats
'''



import os
import socket
import subprocess
import sys
import yodel.globaldat as globaldat
from typing import *


def setting_update(setting, value):
    """
    modifies settings based on the contents of pipe

    Args:
    	setting: setting that is to be modified

    	value: what that setting should be set to

    """

    if setting == "name":  # change name
        globaldat.robotName = value
    elif setting == "add_group":  # add to new group
        globaldat.groups.append(value)
    elif setting == "del_group":  # delete group
        if value in globaldat.groups:
            loc = globaldat.groups.index(value)
            del globaldat.groups[loc]
    elif setting == "clr_group":  # clear groups
        globaldat.groups = []
    elif setting == "exit":  # exit
        print("exit sender")

        sys.exit()


def send_to_receiver(data: list) -> NoReturn:
    """some settings changes require data to be sent to a thread this function

    Args:
    	data: list that holds a setting and what value it should be changed to, eg: ["name","tester"]

    """

    globaldat.receiver_pipe.send(data)


def send_to_sender(data):
    """

    some settings changes require data to be sent to sender thread this function takes care of that

    Args:
    	data: list that holds a setting and what value it should be changed to, eg: ["name","tester"]
    """

    globaldat.sender_pipe.send(data)


def setRepeats(num: int) -> NoReturn:
    """ control amount of times a message is repeated during a send

    Args:
    	num: number of time messages should be repeated, larger value means better reliability with reduce bandwidth

    """

    globaldat.totalsends = num


def enableRelay(state: bool) -> NoReturn:
    """ enable relaying of messages to allow extended range for other robots

    Args:
    	state: True:relay on, False: relay off
    """

    globaldat.relay = state


def initPolyNodelYodel():
    enableRelay(True)


def setName(name) -> NoReturn:
    """
    set or change the name of your robot

    Args:
    	name: name of robot
    """

    
    globaldat.robotName = name
    send_to_receiver(["name", name])  # send name change to other threads
    send_to_sender(["name", name])


def getName() -> str:
    """
    returns the name of your robot

    Returns:
        str: name of robot
    """
    return (globaldat.robotName)


def joinGroup(group) -> NoReturn:
    """
    add robot to a new group

    Args:
    	group: name of group to add robot to

    """
    globaldat.groups.append(group)
    send_to_receiver(["add_group", group])
    send_to_sender(["add_group", group])


def leaveGroup(group) -> NoReturn:
    """
    remove robot from specifed group

    Args:
    	group: name of group you want to leave
    """

    if group in globaldat.groups:
        loc = globaldat.groups.index(group)
        del globaldat.groups[loc]
    send_to_receiver(["del_group", group])
    send_to_sender(["del_group", group])


def getGroups() -> list:
    """
    get list of groups robot is a member of

    Returns:
        list: list of group names which this robot is a member of
    """
    return (globaldat.groups)


def clearGroups() -> NoReturn:
    """
    remove robot from all groups
    """
    globaldat.groups = []
    send_to_receiver(["clr_group", False])
    send_to_sender(["clr_group", False])


def setInterface(interface: str) -> NoReturn:
    """
    set the interface used by the robot and initates the socket.

    Args:
    	interface: name of wifi interface that will be used
    """
    globaldat.iface = interface
    globaldat.yodelSocket = socket.socket(
        socket.AF_PACKET,
        socket.SOCK_RAW,
        socket.htons(3))
    globaldat.yodelSocket.bind((interface, 0))


def setChannel(channel:int) -> NoReturn:
    """
    set the channel for the interface, 

    Args:
    	channel: channel number to switch to

    """
    #some drivers only work with iw, some only work with iwconfig so both are included
    os.system(f"sudo iw dev {globaldat.iface} set channel {channel}")
    os.system(f"sudo iwconfig {globaldat.iface} channel {channel}")


def setPower(txdBm:int) -> NoReturn:
    """
    set the transmit power of wifi interface
    3500 is not necessarily a legal or safe power level for your hardware,
    the limiter is just to marginally decrease the odds of causing problems.

    Args:
    	txdBm: transmit power in dBm
    """
    if txdBm > 3500:
        txdBm = 3500
        print("power level is set too high")
    os.system(f"sudo iw dev {globaldat.iface} set txpower fixed {txdBm}")


####
# end of settings managment
####


####
# auto configurator
####

def command(cmd) -> str:
    """
    run a command, get the output
    basic time saver, not generally applicable to all linux cmds

    Args:
    	cmd: command being ran
    """
    cmda = cmd.split(" ")
    print(cmda)
    result = subprocess.run(cmda, stdout=subprocess.PIPE)
    return (result.stdout.decode("utf-8"))


def enableMonitor(interface:str) -> NoReturn:
    """ auto configure monitor mode on interface
        it's ok for this to raise errors

    Args:
    	interface: wifi interface being used
    """
    os.system(f"sudo rfkill unblock wifi; sudo rfkill unblock all")
    os.system(f"nmcli device set {interface} managed no")
    os.system(f"sudo ip link set {interface} down")
    os.system(f"sudo iwconfig {interface} mode Monitor")
    os.system(f"sudo ip link set {interface} up")


def isMonitor(interface:str):
    """
    get if interface is in monitor mode, is lazy but only has to run a couple times and only during setup
    (re-work with ioctl)

    Args:
        interface: interface being used
    """
    a = (command(f"iwconfig {interface}"))
    mode = (a.find("Mode:Monitor"))
    if mode != -1:
        return (True)


def autoConf(interface:str) -> NoReturn:
    """
    auto configure wifi interface

    Args:
    	interface: name of wifi interface being used

    """

    if not isMonitor(interface):  # first try to set it to monitor mode
        enableMonitor(interface)
    if not isMonitor(interface):  # if set fails let the user know
        print("monitor mode is not available, please select a different interface or try a new driver")



import subprocess
import random
import os
import time
import socket
import sys
import threading
from queue import LifoQueue
import multiprocessing as mp
import time

import yodel.globaldat as globaldat
from yodel.config import *
from yodel.classes import *
from .dynamicheaders import *
import yodel.standardformats as standardformats
from .sender import send, sender, sendData, sender_pipe_output
from .receiver import listen, receiver, receiver_pipe_output, incoming
from typing import *

import atexit


@atexit.register
def on_exit() -> NoReturn:
    """
     was having issues getting built in daemon threads working, this is simple fix, also allows for more advanced exit behavior
    """
    # give threads a moment to finish up what they are doing
    time.sleep(globaldat.SOCKET_CLOSE_TIME)
    send_to_receiver(["exit", 0])
    send_to_sender(["exit", 0])


# iface="wlp3s0"
# nmcli device set wlp3s0 managed no
# sudo ip link set wlp3s0 down
# sudo iwconfig wlp3s0 mode Monitor
# sudo ip link set wlp3s0 up
# sudo iw dev wlp3s0 set channel 1 HT20
# nmcli device set wlp3s0 managed no && sudo ip link set wlp3s0 down &&
# sudo iwconfig wlp3s0 mode Monitor && sudo ip link set wlp3s0 up


####
# thread setup and managment
####


def setupThreads() -> NoReturn:
    """
    setup sender and receiver threads

    """
    global sender_thread, receiver_thread, outgoing, incoming

    sender_thread = mp.Process(
        target=sender,
        args=(
            globaldat.outgoing,
            sender_pipe_output,
        ))
    sender_thread.daemon = True
    sender_thread.start()

    receiver_thread = mp.Process(
        target=receiver, args=(
            incoming, receiver_pipe_output,))
    receiver_thread.daemon = True
    receiver_thread.start()


def startRadio(interface: str) -> NoReturn:  # all functions needed to initiate radios
    """
    automatically setups wireless interface to work with yodel. also setups up sender and receiver threads.

    Args:
    	 interface: interface to be used with yodel
    """
    autoConf(interface)
    setInterface(interface)
    setupThreads()


def restartThreads() -> NoReturn:
    """
    used to restart the sender and receiver threads if something breaks
    """
    send_to_receiver(["exit", 0])
    send_to_sender(["exit", 0])
    globaldat.receiver_pipe, receiver_pipe_output = mp.Pipe()
    globaldat.sender_pipe, sender_pipe_output = mp.Pipe()
    setupThreads()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yodelnet", # Replace with your own username
    version="0.0.9",
    author="Alden Quigley",
    author_email="aldenaquigley@gmail.com",
    description="Yodel is a python library that uses WIFI hardware for remote control purposes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aldenq/Yodel",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux"
    ],
    python_requires='>=3.6',
)



import copy
import math
from typing import *

import yodel.globaldat as globaldat


def typeManagment(data: Any) -> bytearray:
    '''
    Take in any type, and turn it into a bytearray
    '''
    dtype = type(data)
    if dtype == str:
        return(bytearray(data.encode(encoding='UTF-8', errors='strict')))
    elif dtype == bytes:
        return(data)


class Flags:
    '''
    Class meant to be used in fields, is an array of bools, used to store flags about the packet.

    Args:
        lookup_table: list of strings used to map keys to bits


    
    '''
    

    def __init__(self, lookup_table:list):
        length = 1
        self.data = [0, 0, 0, 0, 0, 0, 0, 0]
        self.lookup = {}  # lookup table, maps names provided by the field onto indexes in data
        self.a = 2
        if lookup_table:  # check if lookup table is provided
            # index lookup table, check to see if a name has been provided, if
            # so create an entry in lookup dict with the key as the name and
            # the value as the index
            for i in range(len(lookup_table)):
                key = lookup_table[i]
                # checks to see if the name provided for a given matrix is
                # None, this is so that ["a",None,"b"] will only set a key for
                # index 0 and 2
                if key is not None:
                    # create dict entry with key being a name provided and the
                    # value being the index being mapped to
                    self.lookup[key] = i

    def __setitem__(self, key, value):
        if type(key) == str:
            self.data[self.lookup[key]] = value
        else:
            self.data[key] = int(value)

    def __getitem__(self, key):
        if type(key) == str:
            return(self.data[self.lookup[key]])
        return(self.data[key])

    def __bytes__(self):
        out = 0
        for i in range(8):

            val = self.data[7 - i]
            out += val * 2**i
        return(out.to_bytes(1, 'little'))
        # return(int(data,2))

    def __repr__(self):
        out = ''
        for i in range(8):
            val = str(int(self.data[i]))
            out += val
        return(out)


class Format:
    """
    formats are used to store the information needed to encode or decode data.
    eg: first 3 bytes are a string, next 5 are for an int, etc.

    Args:
    	 fields: list of field objects that will define the format
    
    	 mtype: short for message type, allows unique identifiers to be given to your format that will be sent along with the format allowing for the receiver to know what format to use to decode the message
    """


    
    
    
    

    def __init__(self, fields:List, mtype: int = 0):
        self.mtype: int = mtype  # kwargs.get("mtype", 0) #get message type
        supported_types = [int, str, bytearray, Flags]
        # dictionary that holds field data formated as field name: field value
        self.fields_dict = {}
        # fields holds the list of fields provided, still holds lots of useful
        # meta data so it is kept around
        self.fields: list = fields
        self.output: dict = {}  # dict that holds field names and values, this is so that sections on init can just copy the info from here rather than regenerating it

        if self.mtype != 0:  # when a format is created and the message type is not zero store it in the array of message types so that autoDecode can use the format
            globaldat.messages_types[self.mtype] = self

        for i in range(
                len(fields)):  # copy data over and init output with field names
            fname = fields[i].name
            self.output[fname] = 0
            self.fields_dict[fname] = fields[i]

        # self.gen_data()


class Section:
    """
    sections are used to store data and the meta-data needed to encode that data.
    to get extract all of the data in a section use:
    section.fields

    sections can be encoded by using bytes(section), also,
    if a section is used in yodel.send it will automatically handle it.

    Args:
        format: format object to be used when encoding this section
    """






    def __init__(self, format: Format):
        # store format so that it can be accessed later as necessary
        self.__dict__["format"] = format
        # copy empty dict from format which has names already set
        self.__dict__["fields"] = copy.copy(format.output)
        # holds anything that comes after all fields have been filled
        self.__dict__["payload"] = b''

    def print(self):  # fancy print

        type_lookup = {
            bytearray: "Bytearray",
            int: "Int",
            Flags: "Flags",
            bytes: "bytes",
            str: "String"
        }

        for i in list(self.fields.keys()):
            name_len = len(str(i))
            space = 20

            dat_len = len(str(self.fields[i]))
            space2 = 20
            field_type = self.__dict__["format"].fields_dict[i].type

            print_type = type_lookup[field_type]
            if dat_len < space2:
                space2 = space2 - dat_len
            if name_len < space:
                space = space - name_len

            if field_type == str:
                # print rules for strings
                print(
                    f"{i}:{' '*space}\"{self.fields[i]}\"{' '*(space2 - 2)}{print_type}")
            elif field_type == int:
                # print rules for ints
                print(
                    f"{i}:{' '*space}{self.fields[i]}{' '*(space2)}{print_type}")
            elif field_type == Flags:
                # print rules for flags
                print(
                    f"{i}:{' '*space}{self.fields[i]}{' '*(space2)}{print_type}     {list(self.fields[i].lookup.keys())}")
            elif field_type == bytearray:
                print(
                    f"{i}:{' '*space}{self.fields[i]}{' '*(space2)}{print_type}")
        print(f"payload:{' '*space}{self.payload}")

    def __bytes__(self):
        return(evalBytes(self.__dict__["fields"], self.__dict__["format"], self.__dict__["payload"]))

    def __setattr__(self, name, value):
        if name != "payload":
            self.fields[name] = value
        else:
            self.__dict__["payload"] = value

    def __getattr__(self, name):
        if name != "payload":
            return(self.fields[name])
        else:
            return(self.__dict__["payload"])

    def __setitem__(self, key, value):
        self.fields[key] = value

    def __getitem__(self, key):
        return(self.fields[key])

    def __str__(self):
        return(str(self.fields))


class Field:
    '''
    A field is a section of memory meant to hold one value

    Args:
        name: name of field
        _type: data type to use in field
        bytes: when applicable this can hold the length of the field
    '''
    supported_types = [int, str, bytearray, Flags]
    def __init__(self, name: str, _type: Type, *args, bytes=0, min=0, max=0):
        bytes_len: int = bytes

        if _type == int:

            self.min = min
            self.max = max
            if bytes_len:
                self.len = bytes_len
                # signed integers are encoded using sign and magnitude
                self.min = -1 * 2**((bytes_len * 8) - 1)
                self.max = 2**((bytes_len * 8) - 1) - 1
            else:
                # when type is an int len tells us the amount of bits needed to
                # represent the possble options. when type is a str len tells
                # us the amount of bits needed to store the length of the
                # string
                self.len = math.ceil((max - min).bit_length() / 8)
            #self.len  =4
        elif _type == str or _type == bytearray:
            if bytes_len:
                max = bytes_len
            self.min = min
            self.max = max

            self.len = math.ceil((max - min).bit_length() / 8)

        elif _type == Flags:
            self.min = 0
            self.max = 0
            self.len = 1  # flags type is always one byte long
            if len(args) == 1:
                # take the array that holds the bit names
                self.lookup = args[0]
            else:
                self.lookup = False

        self.name = name  # field name
        self.type = _type  # field data type
        # self.len = math.ceil((Max-Min).bit_length()/8) #when type is an in
        # len tells us the amount of bits needed to represent the possble
        # options. when type is a str len tells us the amount of bits needed to
        # store the length of the string


def decode(data: bytearray, encoding: Format) -> Section:
    '''
    Returns list of all field names

    Args:
        data: bytearray of data that you want to decode

        encoding: format object to be used as the decoding rules
    '''
    fnames = list(encoding.fields_dict.keys())
    output = Section(encoding)  # generate new section object to store output
    cpos = 0  # current position, sort of a pointer to the bytearray
    for field in range(len(fnames)):
        fname = fnames[field]  # field name
        fieldobj = encoding.fields_dict[fname]
        ftype = fieldobj.type  # data type of the field
        flen = fieldobj.len  # field length
        # take the next n bytes where n is the length of the field
        fdata = data[cpos:cpos + flen]
        cpos += flen  # incriment the current position by the length of the field
        fmin = fieldobj.min  # min field value

        # all data types need their own custom decoding scheme

        if ftype == str:
            # get the size of the string by taking the first flen bytes and
            # converting them to an int
            strlen = globaldat.getInt(fdata)
            strlen += fmin
            # return the next n bytes where n is the length of the string
            # defined by strlen
            strdat = data[cpos:cpos + strlen]
            cpos += strlen  # move current position forward by the length of the string
            output[fname] = strdat.decode("utf-8")  # decode bytes as utf-8

        elif ftype == bytearray:
            # get the size of the bytearray by taking the first flen bytes and
            # converting them to an int
            bytelen = globaldat.getInt(fdata)
            bytelen += fmin
            # return the next n bytes where n is the length of the string
            # defined by bytelen
            strdat = data[cpos:cpos + bytelen]
            cpos += bytelen  # move current position by the length of the byte array
            # move the raw bytes into the appropriate place in the class
            output[fname] = strdat

        elif ftype == int:
            fout = globaldat.getInt(fdata)
            fout += fmin
            output[fname] = fout

        elif ftype == Flags:
            output[fname] = Flags([])
            # because this is python the only way to turn a byte into a list of
            # bits is to first convert it to an int, convert it to a utf-8
            # encoded string of bits, and split that list, than convert all the
            # terms into ints and return that as a list
            fout = list(bin(globaldat.getInt(fdata))[2:])
            fout = list(map(int, fout))
            # add appropriate 0 padding depedning on the length
            fout = [0] * (8 - len(fout)) + fout

            output[fname].data = fout

    output.payload = data[cpos:]
    return(output)


def evalBytes(field_dict: dict, format: Format,
              payload: bytearray) -> bytearray:
    '''
    Used in the __bytes__ method in the section class. Is used to return the
    bytes based on the field dict.

    Args:
        field_dict: a dictionary where the keys are field names, and the
        values are the values of those fields

        format: format object to be used when encoding data

        payload: any data that should be added on after the section has been formatted

    '''

    out = b''  # output is bytearray
    for i in format.fields_dict.keys():
        field_data = field_dict[i]

        format_field = format.fields_dict[i]  # take the field from the format
        field_type = format_field.type  # get the expected data type of the field
        #fmax = format_field.max
        # min length(when refering to bytearray or string)/ value(when refering
        # to an int)
        fmin = format_field.min
        flen = format_field.len  # length in bytes of field

        if field_type == type(
                field_data):  # check if the expected data type matches the actual type

            if field_type == int:
                # the amount of bytes for the int is included in the standard
                # so it does not need to be added to the output
                field_data -= format_field.min

                out += field_data.to_bytes(flen, 'little')

            elif field_type == Flags:  # flags are always 1 byte
                out += bytes(field_data)

            elif field_type == str:
                field_len = len(field_data)
                field_len -= fmin  # minimum length is subtracted because the reciever will add the min back before reading the bytes

                # for string the length of the string first needs to be added
                # as an int before the string data
                out += field_len.to_bytes(flen, 'little')

                # strings are encoded as a utf-8 string
                out += bytearray(field_data.encode(encoding='UTF-8',
                                                   errors='strict'))

            elif field_type == bytearray:
                field_len = len(field_data)
                field_len -= fmin
                # like strings, with byte arrays the length is added prior to
                # the data
                out += field_len.to_bytes(flen, 'little')
                out += field_data

            pass
    out += payload
    return(out)


def autoDecode(data: Section) -> Union[Section, bytearray]:
    '''
    automatically decodes the payload section of a given section, for this to work the format must be defined earlier on. 

    Args: 
    	data: Section a section to decode automatically
    @returns either a Section containing the formatted values of 'data', or data's payload
    '''
    mtype = data.mtype
    if mtype != 0:
        byte_data = data.payload
        return(decode(byte_data, globaldat.messages_types[mtype]))
    else:
        return(data.payload)
